(* $Id: lexer.mll,v 1.21 2001/11/18 23:15:43 malekith Exp $ *)

(* Lexer for the gont language. *)

{

open Parser        (* The type token is defined in parser.mli *)
;;

type lexer_location = {mutable ll_line : int;
                       mutable ll_char : int;
                       mutable ll_file : string}
                       
let comment_depth = ref 0
let comment_start = {ll_file = ""; ll_line = 0; ll_char = 0}
let current_location = {ll_file = "stdin"; ll_line = 1; ll_char = 0}
;;

let keywords = Hashtbl.create 50;;

List.iter (fun (s, t) -> Hashtbl.add keywords s t)
    [ 
      ("bool", KW_BOOL);
      ("case", KW_CASE);
      ("do", KW_DO);
      ("else", KW_ELSE);
      ("extern_c", KW_EXTERN_C);
      ("export_as", KW_EXPORT_AS);
      ("float", KW_FLOAT);
      ("for", KW_FOR);
      ("fun", KW_FUN);
      ("if", KW_IF);
      ("int", KW_INT);
      ("let", KW_LET);
      ("return", KW_RETURN);
      ("skip", KW_SKIP);
      ("string", KW_STRING);
      ("swicth", KW_SWITCH);
      ("void", KW_VOID);
      ("while", KW_WHILE)
    ]
;;

let process_id id =
  try Hashtbl.find keywords id with Not_found -> ID id

let line_caches : (string, int list ref) Hashtbl.t = Hashtbl.create 16

(* exported *)
exception Error of string * int * string
exception Eof

let get_file_name () = current_location.ll_file
let set_file_name s = current_location.ll_file <- s
let get_line_column file_name char_no =
  try
    let stack = Hashtbl.find line_caches file_name in
    let rec helper lst =
      match lst with
          x::xs ->
            if char_no >= x then
              (x, (List.length xs) + 1)
            else
              helper xs
        | [] ->
            (0, 1) in
    let (beg, line) = helper !stack in
    (line, char_no - beg + 1)
  with Not_found -> (1, char_no + 1)
  
let get_current_location () =
        (current_location.ll_file,
         current_location.ll_char)
;;
(* /exported *)

let next_line lexbuf = 
  current_location.ll_line <- current_location.ll_line;
  current_location.ll_char <- (Lexing.lexeme_end lexbuf);
  let stack = 
    try
      Hashtbl.find line_caches (get_file_name ())
    with Not_found -> 
      let lst = ref [0] in
        Hashtbl.add line_caches (get_file_name ()) lst; lst
  in
    stack := ((Lexing.lexeme_end lexbuf)) :: !stack
;;

let unescape st c =
  match c with
     '\\' -> 
         (match Stream.next st with
             'n' -> '\n'
           | 'r' -> '\r'
           | 'b' -> '\b'
           | x -> x)
    | x -> x
;;
 
(* Process \'s in string. Discard first and last '"' *)
let process_string s =
  let r = String.create (String.length s) in
  let seen_close = ref false in
  let i = ref 0 in
  let st = Stream.of_string s in
    Stream.junk st;
    Stream.iter 
          (fun c -> 
            if !seen_close then 
              assert false
            else if c == '"' then
              seen_close := true
            else
              (r.[!i] <- unescape st c; incr i))
          st;
     STRING(String.sub r 0 !i)

let make_char s =
  let st = Stream.of_string s in
  CHAR(unescape st (Stream.next st))

;;

}

let nl = ('\r' '\n' | '\n' | '\r')

let num = ['0'-'9']

let id_beg = ['a'-'z''A'-'Z''_']
let id_body = (id_beg | ['0'-'9''\''])

rule token = parse
    [' ' '\t']          { token lexbuf }     (* skip blanks *)
  | nl                  { next_line lexbuf; token lexbuf }
  | ((num+ '.' num+ 'e' (['+' '-']?) num+) | (num+ 'e' (['+' '-']?) num+))
                        { FLOAT(Lexing.lexeme lexbuf) }
  | (num+ | '0' ['x' 'X'] ['0'-'9''a'-'f''A'-'F']+ | '0' ['b' 'B'] ['0' '1']+)
                        { INT(Int64.to_string (Int64.of_string(Lexing.lexeme lexbuf))) }
  | "'" _ "'"           { make_char(String.sub (Lexing.lexeme lexbuf) 1 1) }
  | "'\\" _ "'"         { make_char(String.sub (Lexing.lexeme lexbuf) 1 2) }
  | '"' ([ ^ '"' ] | '\\' [ '\\' '"' ])* '"'
                        { process_string(Lexing.lexeme lexbuf) }
  | id_beg id_body*     { process_id(Lexing.lexeme lexbuf) }
  | "'" id_beg id_body* { let t = Lexing.lexeme lexbuf in 
                          TYVAR(String.sub t 1 ((String.length t) - 1)) }
  | "/*"                { comment_depth := 1;
                          comment_start.ll_file <- current_location.ll_file;
                          comment_start.ll_line <- current_location.ll_line;
                          comment_start.ll_char <- (Lexing.lexeme_start lexbuf);
                          comment lexbuf; 
                          token lexbuf }
  | "//" [^ '\n' '\r']* nl { next_line lexbuf; token lexbuf }
  | "->"                { ARROW }
  | "<<"                { LTLT }
  | ">>"                { GTGT }
  | ">"                 { GT }
  | "<"                 { LT }
  | "<="                { LE }
  | ">="                { GE }
  | "=="                { EQ }
  | "!="                { NE }
  | "+."                { PLUS_DOT }
  | "-."                { MINUS_DOT }
  | "*."                { TIMES_DOT }
  | "/."                { DIV_DOT }
  | '='                 { EQUAL }
  | '%'                 { MOD }
  | '+'                 { PLUS }
  | '-'                 { MINUS }
  | '*'                 { TIMES }
  | '/'                 { DIV }
  | '('                 { LPAREN }
  | ')'                 { RPAREN }
  | '{'                 { LBRACE }
  | '}'                 { RBRACE }
  | ';'                 { SEMICOLON }
  | ','                 { COMMA }
  | "::"                { DOUBLECOLON }
  | ':'                 { COLON }
  | '['                 { LSQUARE }
  | ']'                 { RSQUARE }
  | eof                 { raise Eof }
  | _                   
    { raise (Error(get_file_name(), 
                   (Lexing.lexeme_start lexbuf), 
                   "invalid character in input `" ^ 
                    (String.escaped (Lexing.lexeme lexbuf)) ^ "'")) }

and comment = parse
    "*/"		{ decr comment_depth; 
                          if !comment_depth > 0 then (comment lexbuf) }
  | "/*"                { incr comment_depth; comment lexbuf }
  | nl                  { next_line lexbuf; comment lexbuf }
  | eof                 
   { raise (Error (comment_start.ll_file, 
                   comment_start.ll_line,
                   "EOF within comment")) }
  | _			{ comment lexbuf }

